from __future__ import annotations

import logging
from copy import deepcopy
from dataclasses import dataclass
from datetime import timedelta
from typing import TYPE_CHECKING, NamedTuple

import requests
from skyfield.api import Time, load
from skyfield.units import Angle, Velocity

if TYPE_CHECKING:
    from skyfield.toposlib import GeographicPosition

    from .satellite import Satellite

logger = logging.getLogger(__name__)


class PassEvent(NamedTuple):
    time: Time
    az: Angle
    el: Angle


# While skyfield is capable of array math the interesting events in a pass
# are of low numbers and clarity is more important
# FIXME: could this just be Navigator?
# FIXME: should it know of ts so I can say time_to_rise() -> timedelta
@dataclass
class PassInfo:
    rise: PassEvent
    culm: list[PassEvent]
    fall: PassEvent


class Tracker:
    def __init__(
        self,
        observer: GeographicPosition,
        *,
        owmid: str = '',
    ) -> None:
        '''Tracks a satellite relative to a given observer.

        Given (or told to fetch) a TLE, it can handle temperature and pressure calibration,
        determining when the next pass will be, current satellite location, and doppler scaling.

        Parameters
        ----------
        observer
            The latitude, longitude, and altitude of the observer.
        owmid
            Open Weather Map API key, or empty string for none.
        '''
        self.owmid = owmid
        self.obs = observer
        self.ts = load.timescale()

    def _build_passinfo(self, sat: Satellite, times: list[Time]) -> PassInfo:
        events = []
        for t in times:
            # Not compensating for temperature/pressure because the culm could be well in the future
            # Recomputed later accounting for them in track()
            el, az, _ = (sat - self.obs).at(t).altaz()
            events.append(PassEvent(t, az, el))
        return PassInfo(events[0], events[1:-1], events[-1])

    # FIXME: filtering by time of day
    def next_pass(  # noqa: C901
        self,
        sat: Satellite,
        after: Time | None = None,
        min_el: Angle | None = None,
        lookahead: timedelta = timedelta(days=3),
    ) -> PassInfo | None:
        if after is None:
            after = self.ts.now()

        if min_el is None:
            min_el = Angle(degrees=15)

        # Find all passes above 0°. find_events() does have an elevation argument but that returns
        # times at the given elevation and we want the times where it crosses the horizon, 0°.
        times, events = sat.find_events(self.obs, after, after + lookahead)
        if not times:
            logger.error("No pass found in the next %d", lookahead)
            logger.info("The TLE or station location is fishy.")
            return None

        # Events are the following integers
        RISE = 0  # Satellite rises above 0° # noqa: N806
        CULM = 1  # Satellite is at an elevation local maxima # noqa: N806, F841
        FALL = 2  # Satellite falls below 0° # noqa: N806

        # Split the times list by pass, which is [RISE, one or more CULM, FALL], but we may be in a
        # pass so pre-fill the missing parts for the first entry.
        passes = []
        singlepass = []
        # `after` may be during a pass so we don't have an initial RISE/CULM
        if events[0] != RISE:
            # Some time during a pass
            singlepass.append(after)
        if events[0] == FALL:
            # Past the final culmination so `after` is the max elevation
            singlepass.append(after)

        for t, e in zip(times, events, strict=True):
            singlepass.append(t)
            if e == FALL:
                passes.append(singlepass)
                singlepass = []

        # The pass may end after the look-ahead time, so truncate. Either the orbit is unusual,
        # like a geosynchronous orbit, or the user should wait and recompute with the next
        # available TLE
        # FIXME: this is kind of unsatisfactory, the user may have to recompute during a pass.
        # Is there a better way of handling this?
        if events[-1] == RISE:
            singlepass.append(after + lookahead)
        if events[-1] != FALL:
            singlepass.append(after + lookahead)
            passes.append(singlepass)
            # FIXME: log.warn on truncation, but only if it's the only pass and only if
            # it's above min_el

        # Find the next pass with culmination greater than min_el
        for singlepass in passes:
            info = self._build_passinfo(sat, singlepass)
            for culm in info.culm:
                if culm.el.radians >= min_el.radians:
                    return info

        logger.warning(
            "No passes for %s in the next %s with elevation >=%.1f°",
            sat.name,
            lookahead,
            min_el.degrees,
        )
        return None

    def weather(self) -> tuple[float, float]:
        if not self.owmid:
            # From the ephem docs, temperature defaults to 25 C, pressure defaults to 1010 mBar
            logger.info("not fetching weather for calibration")
            return (25.0, 1010.0)
        # See https://openweathermap.org/api/one-call-3
        # FIXME: Rate limit 1 per mintue/1000 per day
        r = requests.get(
            'https://api.openweathermap.org/data/3.0/onecall',
            params={
                'lat': f'{self.obs.latitude.degrees:.3f}',
                'lon': f'{self.obs.longitude.degrees:.3f}',
                'exclude': 'minutely,hourly,daily,alerts',
                'units': 'metric',
                'appid': self.owmid,
            },
            timeout=10,
        )
        r.raise_for_status()
        logger.debug("Weather response: %s", r.json())
        c = r.json()["current"]
        return (c['temp'], c['pressure'])

    def track(
        self, sat: Satellite, singlepass: PassInfo, temp: float = 25.0, pressure: float = 1010.0
    ) -> tuple[tuple[Time, Angle, Angle], tuple[Time, Velocity]]:
        """Create a track of a given satellite relative to the set observer.

        Note that as the satellite is approaching the observer the rangevel is negative, and as it's
        moving away the value is positive.
        """
        rise = singlepass.rise.time
        fall = singlepass.fall.time

        pass_duration_seconds = (fall - rise) * 24 * 60 * 60
        passtimes = self.ts.linspace(rise, fall, int(pass_duration_seconds / 2))
        positions = (sat - self.obs).at(passtimes)

        el, az, _ = positions.altaz(temperature_C=temp, pressure_mbar=pressure)
        rangevel = positions.frame_latlon_and_rates(self.obs)[-1]
        # FIXME: separate functions for azel/rangevel?
        # Ideally the radio and the rotator should figure out the timesteps they need
        # Really they should have uniform steps in their required values and time should
        # be the variable but that's not the expected way - hard to do.
        return ((passtimes, az, el), (deepcopy(passtimes), rangevel))



from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from math import cos, pi
from math import degrees as deg
from typing import TYPE_CHECKING

import numpy as np
from skyfield.units import Angle

if TYPE_CHECKING:
    from .tracker import PassInfo

logger = logging.getLogger(__name__)


class Navigator(metaclass=ABCMeta):
    def __init__(self, info: PassInfo) -> None:
        '''Determine rotator position from a given pass.

        As the satellite tracks across the sky, its specific path might exceed
        the slew rate of the rotator, or move beyond the bounds that the
        rotator can reach. There are multiple rotator orientations that point
        at the same spot in the sky though so this class tries to choose the
        correct orientation that keeps the rotator within its limits.

        The three navigation modes currently implemented:
        - normal: The rotator points at the track in a 1-to-1 manner
        - backhand: If az crosses from 0 to 360 this mode inverts the az and
              mirrors the el, so that az should pass through the other side
        - flip: If el is too high, crossing overhead will cause az to traverse
              too quickly. This fixes az at the middle point and moves only el

        Parameters
        ----------
        info
            The specific satellite pass to determine navigation mode from
        '''
        self.info = info

    @classmethod
    def mode(cls, info: PassInfo) -> Navigator:
        # FIXME: This assumes a single culmination
        # Max elevation degrees above which to use nav mode flip.
        # FIXME: This should be calculated from slew rate.
        flip_above = 78

        if info.culm[0].el.degrees >= flip_above:
            return Flip(info)
        if cls.no_zero_cross(info.rise.az.radians, info.culm[0].az.radians, info.fall.az.radians):
            return Straight(info)
        if cls.no_zero_cross(
            *cls.rot_pi((info.rise.az.radians, info.culm[0].az.radians, info.fall.az.radians))
        ):
            return Backhand(info)
        # FIXME This probably means we need to extend into the 450°
        # operating area
        logger.warning("Path and inverse both cross zero, defaulting to straight")
        return Straight(info)

    def __str__(self) -> str:
        rise = self.info.rise.az.radians
        culm = self.info.culm[0].az.radians
        fall = self.info.fall.az.radians

        rise_time = self.info.rise.time.astimezone(None)
        culm_time = self.info.culm[0].time.astimezone(None)
        fall_time = self.info.fall.time.astimezone(None)

        z = (abs(rise - culm) + abs(culm - fall)) > (1.5 * pi)
        return ''.join(
            f"rise: {rise_time} {deg(rise):.1f}°az\n"
            f"culm: {culm_time} {self.info.culm[0].el.degrees:.1f}°el\n"
            f"fall: {fall_time} {deg(fall):.1f}°az\n"
            f"Zero_cross: {z} mode: {self.__class__.__name__}",
        )

    @staticmethod
    def rot_pi(rad: tuple[float, ...]) -> tuple[float, ...]:
        """Rotate any radian by half a circle."""
        return tuple((x + pi) % (2 * pi) for x in rad)

    @staticmethod
    def no_zero_cross(a: float, b: float, c: float) -> bool:
        return (a < b < c) or (a > b > c)

    @staticmethod
    def az_n_hem(az: float) -> bool:
        return cos(az) > 0

    @abstractmethod
    def azel(self, az: Angle, el: Angle) -> tuple[Angle, Angle]:
        pass


class Straight(Navigator):
    def azel(self, az: Angle, el: Angle) -> tuple[Angle, Angle]:
        return (az, el)


class Backhand(Navigator):
    def azel(self, az: Angle, el: Angle) -> tuple[Angle, Angle]:
        return (Angle(radians=(az.radians + pi) % (2 * pi)), Angle(radians=pi - el.radians))


class Flip(Navigator):
    def __init__(self, info: PassInfo) -> None:
        '''Navigation strategy that primarily rotates el and tries to minimize az.

        This prevents the wrist flip problem in high angle passes.

        Parameters
        ----------
        info
            The parameters of the pass to plan for
        '''
        super().__init__(info)
        # halfway point between rise and (fall + halfcircle)
        # FIXME: verify average works on circle
        flip_az = ((info.rise.az.radians + info.fall.az.radians + pi) / 2) % (2 * pi)
        # FIXME: I can't convince myself that this is correct yet
        if self.az_n_hem(flip_az):
            (flip_az,) = self.rot_pi((flip_az,))

        self.flip_az = Angle(radians=flip_az)

    def azel(self, az: Angle, el: Angle) -> tuple[Angle, Angle]:
        # A flip only moves elevation so azimuth is fixed at flip_az and elevation needs to be
        # extended across 0 - 180, passing through 90. Since the target path generally won't follow
        # the navigated path (e.g. has a max el of 80, whereas this path must always go through 90)
        # we need to pick the el at each step that minimizes the error between antenna and target.
        # FIXME: Still working on reasoning this out. For example I'd expect at 0 target el this
        #        would set 0 antenna el, but it doesn't.
        flip_el = pi / 2 - (np.cos(az.radians - self.flip_az.radians) * (pi / 2 - el.radians))
        return (Angle(radians=np.full(len(flip_el), self.flip_az.radians)), Angle(radians=flip_el))

    def __str__(self) -> str:
        val = super().__str__()
        val += f" at {self.flip_az.degrees:.1f}°az"
        return val



import matplotlib.pyplot as plt
import numpy as np

min_el = 0
max_el = 180


# check real az/el
# chack gradient

# feel like with the right range for az/el the special operation could just be negation?

az, el = np.meshgrid(np.linspace(-180, 180, 90), np.linspace(-90, 90, 45))
raz = np.where(el < 0, (az + 360) % 360 - 180, az)
rel = np.where(el < 0, -el, el)


fig, ax = plt.subplots()
ax.quiver(az, el, raz, rel)
plt.show()



from dataclasses import InitVar, dataclass, field
from ipaddress import AddressValueError, IPv4Address
from numbers import Real
from pathlib import Path, PosixPath
from socket import gaierror, gethostbyname
from typing import Any, NamedTuple, TypeAlias

import tomlkit
from sgp4 import earth_gravity, io
from skyfield.api import E, N, wgs84
from skyfield.toposlib import GeographicPosition
from skyfield.units import Angle
from tomlkit.items import Table
from tomlkit.toml_document import TOMLDocument


class ConfigError(Exception):
    pass


class TleValidationError(ConfigError):
    def __init__(self, name: str, tle: list[str]) -> None:  # noqa: D107
        super().__init__(f'TLE for {name}')
        self.name = name
        self.tle = tle


class IpValidationError(ConfigError):
    def __init__(self, table: str | None, key: str, value: Any) -> None:  # noqa: D107 ANN401
        super().__init__(f"'{table}.{key}'")
        self.table = table
        self.key = key
        self.value = value


class KeyValidationError(ConfigError):
    def __init__(self, table: str | None, key: str, expect: str, actual: str) -> None:  # noqa: D107
        super().__init__(f"'{key}' invalid type {actual}")
        self.table = table
        self.key = key
        self.expect = expect
        self.actual = actual


class AngleValidationError(ConfigError):
    def __init__(self, table: str | None, key: str, value: Any) -> None:  # noqa: D107 ANN401
        super().__init__(f"'{table}.{key}'")
        self.table = table
        self.key = key
        self.value = value


class TemplateTextError(ConfigError):
    def __init__(self, table: str | None, key: str) -> None:  # noqa: D107
        super().__init__(f'{table}.{key}')
        self.table = table
        self.key = key


class UnknownKeyError(ConfigError):
    def __init__(self, keys: list[str]) -> None:  # noqa: D107
        super().__init__(' '.join(keys))
        self.keys = keys


class MissingKeyError(ConfigError):
    def __init__(self, table: str | None, key: str) -> None:  # noqa: D107
        super().__init__(f'{table}.{key}')
        self.table = table
        self.key = key


class MissingTableError(ConfigError):
    def __init__(self, table: str) -> None:  # noqa: D107
        super().__init__(table)
        self.table = table


class InvalidTomlError(ConfigError):
    pass


class ConfigNotFoundError(ConfigError):
    pass


class AzEl(NamedTuple):
    az: Any
    el: Any


TleCache: TypeAlias = dict[str, list[str]]
_marker = object()  # marks no default value, in case we want None as default


def _pop_table(cfg: TOMLDocument, table: str, default: Any = _marker) -> Any:  # noqa: ANN401
    try:
        entry = cfg.pop(table)
        if not isinstance(entry, Table):
            raise UnknownKeyError([table])
    except tomlkit.exceptions.NonExistentKey as e:
        if default is _marker:
            raise MissingTableError(table) from e
        entry = default
    return entry


def _pop(table: Table, key: str, valtype: type, default: Any = _marker) -> Any:  # noqa: ANN401
    try:
        val = table.pop(key)
    except tomlkit.exceptions.NonExistentKey as e:
        if default is _marker:
            raise MissingKeyError(table.display_name, key) from e
        val = default
    if not isinstance(val, valtype):
        raise KeyValidationError(table.display_name, key, valtype.__name__, type(val).__name__)
    return val


def _pop_ip(table: Table, key: str, valtype: type, default: Any = _marker) -> IPv4Address:  # noqa: ANN401
    value = _pop(table, key, valtype, default)
    try:
        return IPv4Address(gethostbyname(value))
    except (AddressValueError, gaierror) as e:
        raise IpValidationError(table.display_name, key, value) from e


def _check_template_text(config: TOMLDocument) -> None:
    '''Ensure all template text has been removed.'''
    for name, table in config.items():
        if not isinstance(table, Table):
            continue
        for key, value in table.items():
            if isinstance(value, str) and '<' in value:
                raise TemplateTextError(name, key)


@dataclass
class Config:
    path: InitVar[Path]

    # Main
    sat_id: str = ''
    min_el: Angle = Angle(degrees=15.0)  # noqa: RUF009 __post_init__ will create a copy
    owmid: str = ''
    edl: tuple[str, int] = ('', 10025)
    txgain: int = 2

    # Hosts
    edl_dest: tuple[str, int] = ('127.0.0.1', 10025)
    flowgraph: tuple[str, int] = ('127.0.0.1', 10080)
    station: tuple[str, int] = ('127.0.0.1', 5005)
    rotator: PosixPath = PosixPath('/dev/null')  # noqa: RUF009

    # Observer
    # lat, lon, alt TOML fields. Default value here is Portland ebv1 station.
    observer: GeographicPosition = wgs84.latlon(45.509054 * N, -122.681394 * E, 50)
    name: str = ''
    cal: AzEl = AzEl(0, 0)
    slew: AzEl | None = None
    beam_width: float | None = None
    # FIXME: the limit should be retrieved from stationd instead of our toml but that feature
    #        doesn't exist.
    temp_limit: float = 40.0

    # Satellite
    tle_cache: TleCache = field(default_factory=dict)

    # Command line only
    mock: set[str] = field(default_factory=set)
    pass_count: int = 9999

    def __post_init__(self, path: Path) -> None:
        '''Load a config from a given file.

        Checks:
        - File exists and is valid toml
        - All template text removed
        - Mandatory tables exist and are Tables
        - Optional tables, if they exist, are Tables
        - Mandatory keys exist and have values are the expected toml type
        - Optional keys, if they exist, have values that are the expected toml type
        - Values convert to the expected Config type
        - No unexpected keys/all keys consumed

        Parameters
        ----------
        path
            Path to the config file, usually pass_commander.toml
        '''
        try:
            config = tomlkit.parse(path.expanduser().read_text())
        except tomlkit.exceptions.ParseError as e:
            raise InvalidTomlError(*e.args) from e
        except FileNotFoundError as e:
            raise ConfigNotFoundError from e
        except IsADirectoryError as e:
            raise ConfigNotFoundError from e

        _check_template_text(config)

        main = _pop_table(config, 'Main')
        self.sat_id = str(_pop(main, 'satellite', str, self.sat_id))
        self.min_el = Angle(degrees=_pop(main, 'minimum-pass-elevation', Real, self.min_el.degrees))
        self.owmid = str(_pop(main, 'owmid', str, self.owmid))
        self.edl = ('', int(_pop(main, 'edl_port', int, self.edl[1])))
        self.txgain = int(_pop(main, 'txgain', int))

        hosts = _pop_table(config, 'Hosts')
        radio = _pop_ip(hosts, 'radio', str)
        station = _pop_ip(hosts, 'station', str)
        self.edl_dest = (str(radio), self.edl_dest[1])
        self.flowgraph = (str(radio), self.flowgraph[1])
        self.station = (str(station), self.station[1])
        self.rotator = PosixPath(_pop(hosts, 'rotator', str))

        observer = _pop_table(config, 'Observer')
        self.observer = wgs84.latlon(
            _pop(observer, 'lat', Real) * N,
            _pop(observer, 'lon', Real) * E,
            _pop(observer, 'alt', int),
        )
        if not -90 <= self.observer.latitude.degrees <= 90:
            raise AngleValidationError(observer.display_name, 'lat', self.observer.latitude.degrees)
        if not -180 <= self.observer.longitude.degrees <= 180:
            raise AngleValidationError(
                observer.display_name, 'lon', self.observer.longitude.degrees
            )
        self.name = str(_pop(observer, 'name', str))  # XMLRPC can't handle toml subclass
        self.temp_limit = float(_pop(observer, 'temperature-limit', Real, self.temp_limit))

        self.tle_cache = dict(_pop_table(config, 'TleCache', {}))
        # validate TLEs
        for key, tle in self.tle_cache.items():
            try:
                io.twoline2rv(tle[1], tle[2], earth_gravity.wgs84)
            except (IndexError, ValueError) as e:
                raise TleValidationError(key, tle) from e

        # Ensure there's no extra keys
        extra = ['Main.' + k for k in main]
        extra.extend('Hosts.' + k for k in hosts)
        extra.extend('Observer.' + k for k in observer)
        extra.extend(k for k in config)
        if extra:
            raise UnknownKeyError(extra)

    @classmethod
    def template(cls, path: Path) -> None:
        config = tomlkit.document()
        config.add(tomlkit.comment("Be sure to replace all <hint text> including angle brackets"))
        config.add(tomlkit.comment("Optional fields are commented out, uncomment to set"))

        main = tomlkit.table()
        main.add(tomlkit.comment('satellite = "<Index to TleCache, Gpredict, or NORAD ID>"'))
        main.add(tomlkit.comment('owmid = "<OpenWeatherMap API key>"'))
        main['edl_port'] = cls.edl[1]
        main['txgain'] = cls.txgain

        hosts = tomlkit.table()
        hosts['radio'] = str(cls.flowgraph[0])
        hosts['station'] = str(cls.station)
        hosts['rotator'] = str(cls.rotator)

        observer = tomlkit.table()
        observer.add(tomlkit.comment("Change lat, lon, and alt to your specific station."))
        observer.add(tomlkit.comment("These values are for the Portland evb1 station"))
        observer['lat'] = cls.observer.latitude.degrees
        observer['lon'] = cls.observer.longitude.degrees
        observer['alt'] = int(cls.observer.elevation.m)
        observer['name'] = '<station name or callsign>'

        config['Main'] = main
        config['Hosts'] = hosts
        config['Observer'] = observer

        config.add(tomlkit.nl())
        config.add(tomlkit.comment("[TleCache]"))
        config.add(tomlkit.comment("<name> = ["))
        config.add(tomlkit.comment('    "<TLE Title>",'))
        config.add(tomlkit.comment('    "<TLE line 1>",'))
        config.add(tomlkit.comment('    "<TLE line 2>",'))
        config.add(tomlkit.comment("]"))

        path = path.expanduser()
        if path.exists():
            raise FileExistsError

        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(tomlkit.dumps(config))



# ruff: noqa: ERA001, D103
from contextlib import closing
from pathlib import Path

import pytest
import tomlkit
from tomlkit.toml_document import TOMLDocument

from pass_commander import mock
from pass_commander.config import Config
from pass_commander.satellite import Satellite


@pytest.fixture
def good_toml() -> TOMLDocument:
    cfg = tomlkit.document()

    main = tomlkit.table()
    main['satellite'] = "fake-sat"
    main['minimum-pass-elevation'] = 16
    main['owmid'] = "fake-id"
    main['edl_port'] = 12345
    main['txgain'] = 47

    hosts = tomlkit.table()
    hosts['radio'] = "localhost"
    hosts['station'] = "127.0.0.1"
    hosts['rotator'] = "/dev/null"

    observer = tomlkit.table()
    observer['lat'] = 45.509054
    observer['lon'] = -122.681394
    observer['alt'] = 500
    observer['name'] = 'not-real'
    observer['temperature-limit'] = 33.0

    cfg['Main'] = main
    cfg['Hosts'] = hosts
    cfg['Observer'] = observer

    cfg['TleCache'] = {
        'OreSat0': [
            "ORESAT0",
            "1 52017U 22026K   23092.57919752  .00024279  00000+0  10547-2 0  9990",
            "2 52017  97.5109  94.8899 0023022 355.7525   4.3512 15.22051679 58035",
        ],
        '2022-026K': [
            "ORESAT0",
            "1 52017U 22026K   23092.57919752  .00024279  00000+0  10547-2 0  9990",
            "2 52017  97.5109  94.8899 0023022 355.7525   4.3512 15.22051679 58035",
        ],
    }

    return cfg


@pytest.fixture
def good_config(tmp_path: Path, good_toml: TOMLDocument) -> Config:
    path = tmp_path / 'config.toml'
    with path.open('w+') as f:
        tomlkit.dump(good_toml, f)
        f.flush()
    return Config(path)


@pytest.fixture
def stationd() -> tuple[str, int]:
    s = mock.Stationd()
    s.start()
    try:
        yield s.addr
    finally:
        s.close()
        s.join()


@pytest.fixture(params=(1, 2, 4))
def rotator(request) -> str:  # noqa: ANN001
    with closing(mock.PtyRotator(pulses_per_degree=request.param)) as r:
        yield r.client_path


@pytest.fixture
def flowgraph() -> tuple[str, int]:
    f = mock.Flowgraph()
    f.start()
    try:
        yield f
    finally:
        f.close()
        f._thread.join()  # noqa: SLF001


@pytest.fixture
def edl() -> tuple[str, int]:
    e = mock.Edl()
    e.start()
    try:
        yield e
    finally:
        e.close()
        e.join()


tles = [
    # Geostationary
    # [
    #     "GOES 18",
    #     "1 51850U 22021A   25073.61632166  .00000091  00000+0  00000+0 0  9992",
    #     "2 51850   0.0458 328.8804 0000426 100.9638 187.4828  1.00272874 11191",
    # ],
    # Sun synchronous
    [
        "ORESAT0",
        "1 52017U 22026K   24237.61773939  .00250196  00000+0  18531-2 0  9992",
        "2 52017  97.4861 255.7395 0002474 307.8296  52.2743 15.72168729136382",
    ],
]


@pytest.fixture(params=tles)
def sat(request) -> Satellite:  # noqa: ANN001
    name = request.param[0]
    return Satellite(name, tle_cache={name: request.param}, local_only=True)



import logging
import socket
from threading import Lock
from time import sleep
from xmlrpc.client import ServerProxy

from skyfield import constants

logger = logging.getLogger(__name__)


class Radio:
    def __init__(
        self, flowgraph: tuple[str, int], edl: tuple[str, int], name: str, morse_delay: float = 4.0
    ) -> None:
        '''Binding to the uniclogs-sdr GNURadio flowgraph.

        The flowgraph exposes an xmlrpc interface for configuration and setting doppler offsets.
        It additionally has a UDP port for EDL communication.

        Parameters
        ----------
        flowgraph
            IP and port of the flowgraph xmlrpc server
        edl
            IP and port of the EDL UDP socket
        name
            station identifier to be sent out over morse
        morse_delay
            duration in seconds to wait for the morse identifier broadcast to complete
        '''
        self._edl = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._edl.connect(edl)
        self._lock = Lock()
        self._flowgraph = ServerProxy(f"http://{flowgraph[0]}:{flowgraph[1]}")
        tx = self._flowgraph.get_tx_center_frequency()
        rx = self._flowgraph.get_rx_target_frequency()
        if not isinstance(tx, float):
            raise TypeError("Flowgraph returned invalid tx type")
        if not isinstance(rx, float):
            raise TypeError("Flowgraph returned invalid rx type")
        self.txfreq = tx
        self.rxfreq = rx
        self.name = name
        # FIXME: infer default delay from name
        self.morse_delay = morse_delay

    def ident(self) -> None:
        old_selector = self.get_tx_selector()
        self.set_tx_selector("morse")
        self.set_morse_ident(self.name)
        sleep(self.morse_delay)
        self.set_tx_selector(old_selector)

    def rx_frequency(self, range_velocity: float) -> float:
        # RX on the ground frequency. range_velocity is the satellite relative velocity, it
        # starts negative and goes positive so the frequency starts high and goes low
        return float(1 - range_velocity / constants.C) * self.rxfreq

    def set_rx_frequency(self, range_velocity: float) -> None:
        freq = self.rx_frequency(range_velocity)
        logger.info("Set RX frequency %.1f", freq)
        with self._lock:
            self._flowgraph.set_gpredict_rx_frequency(freq)

    def tx_frequency(self, range_velocity: float) -> float:
        # TX is the opposite of RX, starts low, goes high
        return float(1 + range_velocity / constants.C) * self.txfreq

    def set_tx_frequency(self, range_velocity: float) -> None:
        freq = self.tx_frequency(range_velocity)
        logger.info("Set TX frequency %.1f", freq)
        with self._lock:
            self._flowgraph.set_gpredict_tx_frequency(freq)

    def set_tx_selector(self, mode: str) -> None:
        logger.info("Selecting mode %s", mode)
        with self._lock:
            self._flowgraph.set_tx_selector(mode)

    def get_tx_selector(self) -> str:
        with self._lock:
            val = self._flowgraph.get_tx_selector()
            if not isinstance(val, str):
                raise TypeError("Flowgraph returned invalid tx_selector type")
            return val

    def set_tx_gain(self, gain: int) -> None:
        logger.info("Setting gain %s", gain)
        with self._lock:
            self._flowgraph.set_tx_gain(gain)

    def set_morse_ident(self, ident: str) -> None:
        logger.info("Sending morse ident %s", ident)
        with self._lock:
            self._flowgraph.set_morse_ident(ident)

    def edl(self, packet: bytes, range_velocity: float) -> None:
        self.set_tx_frequency(range_velocity)
        self._edl.send(packet)

    def close(self) -> None:
        # FIXME: close _flowgraph?
        self._edl.close()



import logging
import re
import socket
from time import sleep

logger = logging.getLogger(__name__)


class StationError(Exception):
    pass


class Station:
    def __init__(self, addr: tuple[str, int], band: str = "l-band", lna_delay: float = 1.0) -> None:
        '''Python binding for uniclogs-stationd.

        Yes stationd is written in python but it only exposes a socket interface. Except also
        for reasons this is incomplete and you'll need one instance per band.

        Parameters
        ----------
        addr
            IP address and port of the stationd instance.
        band
            One of `l-band` or `uhf`. We need separate instances to control the radios for each
            band.
        lna_delay
            Time in seconds to wait after toggling the LNA relay. The lna_{on,off} methods toggle
            the relay three times so they'll take 3 * lna_delay seconds.
        '''
        self.s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.s.connect(addr)
        self.band = band
        self.lna_delay = lna_delay

    def _command(self, verb: str) -> str:
        if re.match(
            r"^(gettemp|((l-band|uhf) (pa-power|rf-ptt|lna)|rotator) (on|off|status))$",
            verb,
        ):
            logger.info("Sending command: %s", verb)
            self.s.send(verb.encode())
            return self._response()
        raise StationError(f"invalid command: {verb}")

    def _response(self) -> str:
        # FIXME: timeout, stationd should respond quickly
        data = self.s.recv(4096)
        stat = data.decode().strip()
        logger.info("StationD response: %s", stat)
        if stat.upper().startswith("FAIL: "):
            raise StationError(stat.removeprefix('FAIL: '))
        return stat

    def pa_on(self) -> None:
        self._command(f"{self.band} pa-power on")
        self._command(f"{self.band} pa-power on")

    def pa_off(self) -> None:
        ret = self._command(f"{self.band} pa-power off")
        if m := re.search(r"Please wait (\S+) seconds", ret):
            sleep(int(m.group(1)))
            self.pa_off()

    def ptt_on(self) -> None:
        self._command(f"{self.band} rf-ptt on")
        # FIXME TIMING: wait for PTT to open (100ms is just a guess)
        sleep(0.1)

    def ptt_off(self) -> None:
        self._command(f"{self.band} rf-ptt off")

    def lna_on(self) -> None:
        # The LNA Relay is weird, it's not guaranteed to go on on the first try
        # so it must be cycled. From Glenn in oresat-comms, 2024-09-05:
        # Whereas there only needs to be a 100 ms pulse to switch the relay
        # state, I am unsure of how fast the relay can go through multiple
        # state changes. I would suggest as large as one second between
        # commands just to be safe.
        self._command(f"{self.band} lna on")
        sleep(self.lna_delay)
        self._command(f"{self.band} lna off")
        sleep(self.lna_delay)
        self._command(f"{self.band} lna on")
        sleep(self.lna_delay)

    def lna_off(self) -> None:
        self._command(f"{self.band} lna off")
        sleep(self.lna_delay)
        self._command(f"{self.band} lna on")
        sleep(self.lna_delay)
        self._command(f"{self.band} lna off")
        sleep(self.lna_delay)

    def gettemp(self) -> float:
        ret = self._command("gettemp")
        return float(ret[6:].strip())

    def close(self) -> None:
        self.s.close()



import logging
import re
from pathlib import Path

import requests
from skyfield.api import EarthSatellite

from .config import TleCache

logger = logging.getLogger(__name__)


class Satellite(EarthSatellite):  # type: ignore[misc]
    def __init__(
        self, sat_id: str, *, tle_cache: TleCache | None = None, local_only: bool = False
    ) -> None:
        '''Fetch a TLE and build a satellite.

        Ideally fetched close to a pass for an up-to-date TLE.

        ID formats:
        - International Designator/COSPAR ID: 4 digit year, optional dash, 3
            digit launch number, up to 3 letter sequential identifier
            Example: 2024-149BK
        - NORAD Catalog Number: up to 9 digit number (typically 5)
            Example: 60525
        - Name: official name as it appears in the satellite catalog
            Example: ORESAT0.5

        Parameters
        ----------
        sat_id
            The ID of the satellite, either as an International Designator, a NORAD Satellite
            Catalog number, or Catalog satellite name.
        local_only
            If true, do not use the internet for TLE lookup, only tle_cache or Gpredict.
        tle_cache
            A local cache of TLEs for lookup.
        '''
        query = "NAME"
        if re.match(r"^\d{4}-?\d{3}[A-Z]{1,3}$", sat_id.upper()):
            query = "INTDES"
        elif re.match(r"^\d{1,9}$", sat_id):
            query = "CATNR"

        tle = None
        if tle is None and not local_only:
            # see https://celestrak.org/NORAD/documentation/gp-data-formats.php
            # FIXME: 2 hour cache. This must add the data to the TLE cache. Just use skytraq?
            r = requests.get(
                "https://celestrak.org/NORAD/elements/gp.php",
                params={query: sat_id},
                timeout=10,
            )
            r.raise_for_status()
            lines = r.text.splitlines()
            if lines[0] == "No GP data found":
                logger.info("No results for %s at celestrak", sat_id)
            else:
                tle = lines

        if tle is None and tle_cache and sat_id in tle_cache:
            logger.info("using cached TLE")
            tle = tle_cache[sat_id]

        if tle is None and query == "CATNR":
            fname = Path.home() / f'.config/Gpredict/satdata{sat_id}.sat'
            if fname.is_file():
                logger.info("using Gpredict's cached TLE")
                with fname.open(encoding="ascii") as file:
                    lines = file.readlines()[3:6]
                    tle = [line.rstrip().split("=")[1] for line in lines]

        if tle is None:
            logger.info("No matching TLE for %s is available", sat_id)
            raise ValueError(f"Invalid satellite identifier: {sat_id}")

        logger.info("\n".join(tle))

        super().__init__(tle[1], tle[2], tle[0])



#!/usr/bin/env python3

import logging
import os
import selectors
import socket
from argparse import ArgumentParser
from threading import Thread
from xmlrpc.server import SimpleXMLRPCServer

logger = logging.getLogger(__name__)


class Edl(Thread):
    def __init__(self, addr: tuple[str, int] | None = None) -> None:
        '''Thread that simulates an EDL listening connection.

        Parameters
        ----------
        addr
            IP address and port to listen on. Recommend a non-localhost loopback (like
            127.0.0.2) because pass-commander listens on 127.0.0.1.
        '''
        super().__init__(name=self.__class__.__name__, daemon=True)
        if addr is None:
            addr = ('127.0.0.1', 0)
        self._edl = socket.socket(socket.AF_INET, socket.SOCK_DGRAM | socket.SOCK_NONBLOCK)
        self._edl.bind(addr)
        self._addr: tuple[str, int] = self._edl.getsockname()
        self._r, self._w = os.pipe2(os.O_NONBLOCK)

    @property
    def addr(self) -> tuple[str, int]:
        return self._addr

    def _respond(self) -> bool:
        logger.info("EDL %s", self._edl.recv(4096))
        return False

    def run(self) -> None:
        sel = selectors.DefaultSelector()
        sel.register(self._edl, selectors.EVENT_READ, self._respond)
        sel.register(self._r, selectors.EVENT_READ, lambda: True)

        stop = False
        while not stop:
            for key, _ in sel.select():
                if stop := key.data():
                    break

        sel.close()
        self._edl.close()
        os.close(self._r)
        os.close(self._w)
        logger.info("Stopped")

    def close(self) -> None:
        os.write(self._w, b's')


class FlowgraphState:
    def __init__(self) -> None:
        '''Track the internal simulated flowgraph state.'''
        self.morse_ident = ''
        self.tx_selector = "edl"

    def get_tx_center_frequency(self) -> float:
        return 1_265_000_000.0

    def get_rx_target_frequency(self) -> float:
        return 436_500_000.0

    def set_gpredict_tx_frequency(self, value: float) -> None:
        logger.info("TX Freq %f", value)

    def set_gpredict_rx_frequency(self, value: float) -> None:
        logger.info("RX Freq %f", value)

    def set_morse_ident(self, val: str) -> None:
        self.morse_ident = val
        logger.info("Morse ident %s:", val)

    def get_morse_ident(self) -> str:
        return self.morse_ident

    def set_tx_selector(self, val: str) -> None:
        self.tx_selector = val
        logger.info("TX Selector %s", val)

    def get_tx_selector(self) -> str:
        return self.tx_selector

    def set_tx_gain(self, val: int) -> None:
        logger.info("TX Gain %d", val)


class Flowgraph:
    def __init__(self, addr: tuple[str, int] | None = None) -> None:
        '''Simulate xmlrpc flowgraph interface for testing.

        Parameters
        ----------
        addr
            IP address and port to listen on, usually localhost or some loopback.
        '''
        if addr is None:
            addr = ('127.0.0.1', 0)
        self._state = FlowgraphState()
        self._server = SimpleXMLRPCServer(addr, allow_none=True, logRequests=False)
        self._addr: tuple[str, int] = self._server.socket.getsockname()
        self._server.register_instance(self._state)

        self._r, self._w = os.pipe2(os.O_NONBLOCK)
        self._thread = Thread(target=self._run)

    @property
    def addr(self) -> tuple[str, int]:
        return self._addr

    def _handle(self) -> bool:
        self._server.handle_request()
        return False

    def _run(self) -> None:
        # The stock SimpleXMLRPCServer is based on socketserver that has its own serve_forever
        # but the implementation is based on a half second timeout (and there's a note in the
        # source that says pls fix) which makes unit tests slow. We already do the right thing
        # elsewhere so we can just rebuild it here.
        sel = selectors.DefaultSelector()
        sel.register(self._server, selectors.EVENT_READ, self._handle)
        sel.register(self._r, selectors.EVENT_READ, lambda: True)

        stop = False
        while not stop:
            for key, _ in sel.select():
                if stop := key.data():
                    break

        sel.close()
        self._server.server_close()
        os.close(self._r)
        os.close(self._w)
        logger.info("Stopped")

    def start(self) -> None:
        self._thread.start()

    def close(self) -> None:
        os.write(self._w, b's')


if __name__ == '__main__':
    parser = ArgumentParser("Stubbed flowgraph, for simulation purposes")
    parser.add_argument(
        "-o", "--host", default="127.0.0.2", help="address to use, default is %(default)s"
    )
    parser.add_argument(
        "-u",
        "--uplink-edl-port",
        default=10025,
        type=int,
        help="port to send edl packets, default is %(default)s",
    )
    parser.add_argument(
        "-d",
        "--doppler-xmlrpc-port",
        default=10080,
        type=int,
        help="port to connect xmlrpc to, default is %(default)s",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Verbose logging",
    )

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO if args.verbose else logging.WARNING)

    Edl((args.host, args.uplink_edl_port)).start()
    Flowgraph((args.host, args.doppler_xmlrpc_port)).start()



import dataclasses
import typing
from pathlib import Path
from socket import gethostbyname

import pytest
import tomlkit
from tomlkit.toml_document import TOMLDocument

from pass_commander import config
from pass_commander.config import Config


class TestConfig:
    def test_valid(self, tmp_path: Path, good_toml: TOMLDocument) -> None:
        path = tmp_path / 'valid.toml'
        with path.open('w+') as f:
            tomlkit.dump(good_toml, f)
            f.flush()
        conf = Config(path)
        assert conf.sat_id == good_toml['Main']['satellite']
        assert conf.min_el.degrees == good_toml['Main']['minimum-pass-elevation']
        assert conf.owmid == good_toml['Main']['owmid']
        assert conf.edl == ("", good_toml['Main']['edl_port'])
        assert conf.txgain == good_toml['Main']['txgain']
        assert conf.edl_dest == (gethostbyname(good_toml['Hosts']["radio"]), Config.edl_dest[1])
        assert conf.flowgraph == (gethostbyname(good_toml['Hosts']["radio"]), Config.flowgraph[1])
        assert conf.station == (gethostbyname(good_toml['Hosts']['station']), Config.station[1])
        assert conf.rotator == Path(good_toml['Hosts']['rotator'])
        assert conf.observer.latitude.degrees == good_toml['Observer']['lat']
        assert conf.observer.longitude.degrees == good_toml['Observer']['lon']
        assert conf.observer.elevation.m == good_toml['Observer']['alt']
        assert conf.temp_limit == good_toml['Observer']['temperature-limit']
        assert conf.name == good_toml['Observer']['name']
        for field in dataclasses.fields(conf):
            # The toml string class, a subclass of str, was leaking through conf
            # but xmlrpc can only handle actual str. We might as well sanitize all
            # the other fields too.
            assert type(getattr(conf, field.name)) in (
                # run time type checking is kinda hard, it's gotta be one of these right?
                field.type,
                typing.get_origin(field.type),
                *typing.get_args(field.type),
            )
        assert set(conf.tle_cache) == set(good_toml['TleCache'])

        # satellite, minimum-pass-elevation, owmid, edl_port, temperature-limit, TleCache are
        # optional
        del good_toml['Main']['satellite']
        del good_toml['Main']['minimum-pass-elevation']
        del good_toml['Main']['owmid']
        del good_toml['Main']['edl_port']
        del good_toml['TleCache']
        del good_toml['Observer']['temperature-limit']
        path = tmp_path / 'optional.toml'
        with path.open('w+') as f:
            tomlkit.dump(good_toml, f)
            f.flush()
        Config(path)

    def test_missing(self, tmp_path: Path) -> None:
        with pytest.raises(config.ConfigNotFoundError):
            Config(tmp_path / 'missing')

    def test_invalid(self, tmp_path: Path) -> None:
        path = tmp_path / 'invalid.toml'
        with path.open('w+') as f:
            f.write('this is not the file you are looking for')
            f.flush()
        with pytest.raises(config.InvalidTomlError):
            Config(path)

    @pytest.mark.parametrize(
        'entry',
        [
            ('Main', 'fake', 'foo'),
            ('Hosts', 'fake', 'foo'),
            ('Observer', 'fake', 'foo'),
            ('Fake', tomlkit.table()),
            ('Fake', 3),
        ],
    )
    def test_extra_fields(self, tmp_path: Path, good_toml: TOMLDocument, entry: tuple) -> None:
        if len(entry) == 2:
            good_toml[entry[0]] = entry[1]
        elif len(entry) == 3:
            good_toml[entry[0]][entry[1]] = entry[2]
        else:
            raise TypeError

        path = tmp_path / 'extra_field.toml'
        with path.open('w+') as f:
            tomlkit.dump(good_toml, f)
            f.flush()
        with pytest.raises(config.UnknownKeyError):
            Config(path)

    def test_invalid_ip(self, tmp_path: Path, good_toml: TOMLDocument) -> None:
        good_toml['Hosts']['radio'] = 'not an ip'
        path = tmp_path / 'ip.toml'
        with path.open('w+') as f:
            tomlkit.dump(good_toml, f)
            f.flush()
        with pytest.raises(config.IpValidationError):
            Config(path)

    def test_integer_lat_lon(self, tmp_path: Path, good_toml: TOMLDocument) -> None:
        good_toml['Observer']['lat'] = 45
        good_toml['Observer']['lon'] = -122
        path = tmp_path / "intlatlon.toml"
        with path.open('w+') as f:
            tomlkit.dump(good_toml, f)
            f.flush()
        out = Config(path)
        assert out.observer.latitude.degrees == good_toml['Observer']['lat']
        assert out.observer.longitude.degrees == good_toml['Observer']['lon']

    @pytest.mark.parametrize(
        'latlon',
        [
            ('lat', 91.4),
            ('lat', -91.9),
            ('lon', 181.7),
            ('lon', -181.2),
        ],
    )
    def test_invalid_lat_lon(self, tmp_path: Path, good_toml: TOMLDocument, latlon: tuple) -> None:
        good_toml['Observer'][latlon[0]] = latlon[1]

        path = tmp_path / "latlon.toml"
        with path.open('w+') as f:
            tomlkit.dump(good_toml, f)
            f.flush()
        with pytest.raises(config.AngleValidationError):
            Config(path)

    def test_invalid_tle_missing(self, tmp_path: Path, good_toml: TOMLDocument) -> None:
        del good_toml['TleCache']['OreSat0'][2]
        path = tmp_path / "missing.toml"
        with path.open('w+') as f:
            tomlkit.dump(good_toml, f)
            f.flush()
        with pytest.raises(config.TleValidationError) as e:
            Config(path)
        assert isinstance(e.value.__cause__, IndexError)

    def test_invalid_tle_value(self, tmp_path: Path, good_toml: TOMLDocument) -> None:
        good_toml['TleCache']['OreSat0'][1] = "1 52017U 22026K   23092.5791975"
        path = tmp_path / "invalid.toml"
        with path.open('w+') as f:
            tomlkit.dump(good_toml, f)
            f.flush()
        with pytest.raises(config.TleValidationError) as e:
            Config(path)
        assert isinstance(e.value.__cause__, ValueError)

    @pytest.mark.parametrize('edlport', ["12345", 1.2345])
    def test_edl(self, tmp_path: Path, good_toml: TOMLDocument, edlport: str | float) -> None:
        good_toml['Main']['edl_port'] = edlport
        path = tmp_path / "edlport.toml"
        with path.open('w+') as f:
            tomlkit.dump(good_toml, f)
            f.flush()
        with pytest.raises(config.KeyValidationError):
            Config(path)

    def test_template(self, tmp_path: Path) -> None:
        Config.template(tmp_path / "faketemplate.toml")
        with pytest.raises(config.TemplateTextError):
            Config(tmp_path / "faketemplate.toml")

    def test_template_exists(self, tmp_path: Path) -> None:
        conf = tmp_path / "config.toml"
        conf.touch()
        with pytest.raises(FileExistsError):
            Config.template(conf)



import numpy as np
from skyfield.api import load
from skyfield.units import Angle

from pass_commander.navigator import Backhand, Flip, Navigator, Straight
from pass_commander.tracker import PassEvent, PassInfo

# FIXME: nav tracks should always start/end on el 0/180, or possibly AOS/LOS cutoff?


class TestNavigator:
    ts = load.timescale()

    def test_nav_straight(self) -> None:
        info = PassInfo(
            PassEvent(self.ts.now(), Angle(degrees=45), Angle(degrees=0)),
            [PassEvent(self.ts.now(), Angle(degrees=90), Angle(degrees=45))],
            PassEvent(self.ts.now(), Angle(degrees=135), Angle(degrees=0)),
        )

        nav = Navigator.mode(info)
        assert isinstance(nav, Straight)
        az, el = nav.azel(Angle(radians=np.full(2, 45.0)), Angle(radians=np.full(2, 45.0)))
        assert isinstance(az, Angle)
        assert isinstance(el, Angle)
        str(nav)

    def test_nav_backhand(self) -> None:
        info = PassInfo(
            PassEvent(self.ts.now(), Angle(degrees=350), Angle(degrees=0)),
            [PassEvent(self.ts.now(), Angle(degrees=0), Angle(degrees=45))],
            PassEvent(self.ts.now(), Angle(degrees=170), Angle(degrees=0)),
        )

        nav = Navigator.mode(info)
        assert isinstance(nav, Backhand)
        az, el = nav.azel(Angle(radians=np.full(2, 45.0)), Angle(radians=np.full(2, 45.0)))
        assert isinstance(az, Angle)
        assert isinstance(el, Angle)
        str(nav)

    def test_nav_flip(self) -> None:
        info = PassInfo(
            PassEvent(self.ts.now(), Angle(degrees=45), Angle(degrees=0)),
            [PassEvent(self.ts.now(), Angle(degrees=90), Angle(degrees=90))],
            PassEvent(self.ts.now(), Angle(degrees=135), Angle(degrees=0)),
        )

        nav = Navigator.mode(info)
        assert isinstance(nav, Flip)
        az, el = nav.azel(Angle(radians=np.full(2, 45.0)), Angle(radians=np.full(2, 45.0)))
        assert isinstance(az, Angle)
        assert isinstance(el, Angle)
        # FIXME: el is monotonic from 0 to 180 or viceversa
        str(nav)



#!/usr/bin/env python3

import argparse
import logging
from pathlib import Path
from time import sleep, time

from pass_commander.config import AzEl
from pass_commander.rotator import Rotator

# Assuming ebv1 with the rot2prog/RAS setup. We can only talk to the controlloer once every two
# seconds, and in the 13.8v setup it moves at about 2deg/s.

logging.basicConfig(level=logging.INFO)


def block_until(rot: Rotator, pos: AzEl) -> None:
    '''Block until the rotator reaches the given position.'''
    azel = AzEl(-1, -1)
    while azel != pos:
        sleep(2)
        azel = rot.position()
        print(azel)


def measure(rot: Rotator, start: AzEl, end: AzEl, wait: float = 70.0) -> float:
    '''Measure the time the rotator takes to path from start to end.

    wait should be at least 2 to not talk to the controller too quickly
    '''
    # The hardware set limits are 1deg/179deg
    rot.go(start)
    block_until(rot, start)

    rot.go(end)
    tstart = time()
    sleep(wait)  # should take about 90s so wake up well early and start polling

    azel = rot.position()
    if azel == end:
        raise RuntimeError("Slept too long")

    while azel != end:
        print(azel)
        sleep(2)
        azel = rot.position()
    tend = time()
    return tend - tstart


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('path', type=Path, help='rotator /dev path')
    args = parser.parse_args()

    rot = Rotator(args.path)
    print("Parking")
    rot.park()
    block_until(rot, AzEl(180.0, 90.0))

    print("Starting el measurement")
    t = measure(rot, AzEl(180.0, 1.0), AzEl(180.0, 179.0))
    print("el slew:", (179 - 1) / t, 'deg/s')
    t = measure(rot, AzEl(180.0, 179.0), AzEl(180.0, 1.0))
    print("el slew:", (179 - 1) / t, 'deg/s')

    print("Starting az measurement")
    t = measure(rot, AzEl(180.0, 1.0), AzEl(1.0, 1.0))
    print("az slew @ 1 el:", (180 - 1) / t, 'deg/s')
    t = measure(rot, AzEl(1.0, 1.0), AzEl(180.0, 1.0))
    print("az slew @ 1 el:", (180 - 1) / t, 'deg/s')

    print("Starting az measurement post park")
    t = measure(rot, AzEl(180.0, 90.0), AzEl(1.0, 90.0))
    print("az slew @ 90 el:", (180 - 1) / t, 'deg/s')
    t = measure(rot, AzEl(1.0, 90.0), AzEl(180.0, 90.0))
    print("az slew @ 90 el:", (180 - 1) / t, 'deg/s')

    print("Starting az/el measurement")
    # FIXME: Az and El time should be measured separately since they could have different rates.
    t = measure(rot, AzEl(1.0, 1.0), AzEl(179.0, 179.0))
    print("az/el slew:", (179 - 1) / t, 'deg/s')
    t = measure(rot, AzEl(179.0, 179.0), AzEl(1.0, 1.0))
    print("az/el slew:", (179 - 1) / t, 'deg/s')

    print("Starting polling measurement")
    # See if polling through the entire path slows things down
    t = measure(rot, AzEl(1.0, 1.0), AzEl(1.0, 179.0), wait=2.0)
    print("el slew:", (179 - 1) / t, 'deg/s')
    t = measure(rot, AzEl(1.0, 179.0), AzEl(1.0, 1.0), wait=2.0)
    print("el slew:", (179 - 1) / t, 'deg/s')

    rot.park()



'''Contains external runtime dependency mocks.'''

from .flowgraph import Edl, Flowgraph
from .rotator import PtyRotator
from .station import Stationd

__all__ = ["Edl", "Flowgraph", "PtyRotator", "Stationd"]



from __future__ import annotations

import logging
import os
import struct
from threading import Event, Lock, Thread
from typing import TYPE_CHECKING

import rot2prog

from .config import AzEl
from .navigator import Navigator

if TYPE_CHECKING:
    from pathlib import Path

    from skyfield.units import Angle, Time

    from .tracker import PassInfo


logger = logging.getLogger(__name__)


class RotatorError(Exception):
    def __init__(self, arg: str) -> None:
        '''Exceptions raised by Rotator.'''
        super().__init__(arg)


class Bound:
    def __init__(self, lower: float, upper: float) -> None:
        '''Clamp a value value to a given range.'''
        self.lower = lower
        self.upper = upper

    def shift(self, x: float) -> Bound:
        return Bound(self.lower + x, self.upper + x)

    def clamp(self, x: float) -> float:
        return min(max(x, self.lower), self.upper)

    def __contains__(self, item: float) -> bool:
        return self.lower <= item <= self.upper

    def __str__(self) -> str:
        return f'[{self.lower},{self.upper}]'


class Rotator:
    def __init__(self, path: Path, cal: AzEl = AzEl(0, 0), cmd_interval: float = 2.0) -> None:
        '''Monitor and point an antenna.

        The antenna has maximum and minimum az/el and also a minimum rotation
        step size. This iteration is specific to talking to an alfa rot2prog
        controller which has a limitation that it cannot be sent a command more
        than once every two seconds. This isn't enforced here yet.

        Parameters
        ----------
        path
            path to rotator serial device, e.g. /dev/ttyS0
        cal
            Rotator offsets to apply, e.g. if the physical antenna has been turned.
        cmd_interval
            Time to wait inbetween polling the rotator position
        '''
        self.cal = cal  # FIXME: implement
        # FIXME: set from beamwidth/slew rate/actual min step size?
        # do we actually need a min step size?
        # should azel be filtered to min step?
        # should this be a config option?
        self.step = AzEl(5, 3)  # Minimum rotator stepsize

        # FIXME: only open during pass/close after. Also check mode
        self._rot = rot2prog.ROT2Prog(str(path))
        self.ppd = self._rot.get_pulses_per_degree()
        # FIXME: adjusting bounds by 0.1 is a hack to get around floating point
        # noise but it should be below the controller resolution so I don't think
        # it's incorrect. Need to verify behavior on the actual controller
        self._ppd = Bound(round(-1 / self.ppd, 1) - 0.1, round(1 / self.ppd, 1) + 0.1)
        self._rotlock = Lock()
        self.cmd_interval = cmd_interval

        self._stop: Event | None = None
        self._thread: Thread | None = None
        self._r, self._w = os.pipe2(os.O_NONBLOCK)

    @property
    def listener(self) -> int:
        return self._r

    def _events(self, pos: AzEl) -> None:
        # Only runs from go to commanded position
        self._stop = Event()
        last_reported = None
        # It turns out the rot2prog controller shouldn't be talked to more than once every 2
        # seconds otherwise it can potentially be knocked out of calibration by a degree or two.
        while not self._stop.wait(timeout=self.cmd_interval):
            try:
                with self._rotlock:
                    now = AzEl(*self._rot.status())
            except RuntimeError as e:
                # FIXME: what errors does status actually raise?
                # FIXME: write error to _w
                raise RotatorError("Rotator status failed") from e
            if now == last_reported:
                # FIXME: write error to _w
                raise RotatorError("Rotator movement failed")
            last_reported = now

            # standard com:
            # - event when rotator nears target
            # errors:
            # - Rotator not moving when it should be
            # - Serial communication failure

            if now.az in self._ppd.shift(pos.az) and now.el in self._ppd.shift(pos.el):
                os.write(self._w, struct.pack("ff", now.az, now.el))
                self._stop.set()

        self._stop = None
        self._thread = None

    def event(self) -> AzEl:
        return AzEl(*struct.unpack('ff', os.read(self.listener, 8)))

    def limits(self) -> tuple[Bound, Bound]:
        with self._rotlock:
            lim = self._rot.get_limits()
        return (Bound(lim[0], lim[1]), Bound(lim[2], lim[3]))

    def go(self, pos: AzEl) -> None:
        az, el = pos
        logger.info('%-18s%7.3f°az %7.3f°el', "Moving to", az, el)
        with self._rotlock:
            # FIXME: check result
            self._rot.set(az, el)

    def start_polling(self, pos: AzEl) -> None:
        if self._thread is not None:
            raise RuntimeError("Already waiting for movement")
        self._thread = Thread(
            target=self._events, args=(pos,), name=f"Rotator-{pos.az:.1f}-{pos.el:.1f}"
        )
        self._thread.start()

    def position(self) -> AzEl:
        with self._rotlock:
            return AzEl(*self._rot.status())

    def park(self) -> None:
        self.go(AzEl(180, 90))

    def path(self, np: PassInfo, pos: tuple[Time, Angle, Angle]) -> tuple[Time, Angle, Angle]:
        time, az, el = pos
        # sat az/el to rotator movement
        # TODO: filter by stepsize
        # FIXME: error if nav az el exceeds limits
        nav = Navigator.mode(np)
        logger.info("Nav mode:\n%s", nav)
        az, el = nav.azel(az, el)

        return time, az, el

    def close(self) -> None:
        if self._stop is not None:
            self._stop.set()
        if self._thread is not None:
            self._thread.join()
        with self._rotlock:
            self._rot._ser.close()  # noqa: SLF001



#!/usr/bin/env python3

import logging
import os
import selectors
import socket
from threading import Thread

logger = logging.getLogger(__name__)


class Stationd(Thread):
    def __init__(self, addr: tuple[str, int] | None = None, temperature: float = 25.0) -> None:
        '''Thread that locally simulates stationd for testing.

        Parameters
        ----------
        addr
            IP and port to listen with, usually localhost or some loopback
        temperature
            The default temperature to return from gettemp. Can be changed at runtime via
            self.temperature
        '''
        super().__init__(name=self.__class__.__name__, daemon=True)
        if addr is None:
            addr = ('127.0.0.1', 0)
        self._s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM | socket.SOCK_NONBLOCK)
        self._s.bind(addr)
        self._addr: tuple[str, int] = self._s.getsockname()
        self._r, self._w = os.pipe2(os.O_NONBLOCK)
        self.temperature = temperature

    @property
    def addr(self) -> tuple[str, int]:
        return self._addr

    def _action(self, action: str) -> str:
        # The stationd protocol is more complex than I actually want to try to emulate.
        # This is mostly nonsense returns but class Station seems to accept them without
        # crashing.
        words = action.split()
        if len(words) == 1 and words[0] == 'gettemp':
            return f"temp: {self.temperature}"
        if len(words) == 3:
            if words[0] not in self.state:
                return 'FAIL: Invalid Command'
            if words[1] not in self.state[words[0]]:
                return 'FAIL: Invalid Command'

            # Error if ptt is on and
            #  - turning PA off
            #  - turning LNA on
            if words[1] == 'pa-power' and words[2] == 'off' and self.state[words[0]]['rf-ptt']:
                return f'FAIL: {action} PTT Conflict'
            if words[1] == 'lna' and words[2] == 'on' and self.state[words[0]]['rf-ptt']:
                return f'FAIL: {action} PTT Conflict'

            match words[2]:
                case 'on':
                    self.state[words[0]][words[1]] = True
                    return f'SUCCESS: {action}'
                case 'off':
                    self.state[words[0]][words[1]] = False
                    return f'SUCCESS: {action}'
                case 'status':
                    return f'{action} {"ON" if self.state[words[0]][words[1]] else "OFF"}'
                case _:
                    return 'FAIL: Invalid Command'
        return 'FAIL: Invalid Command'

    def _respond(self) -> bool:
        packet, addr = self._s.recvfrom(4096)
        action = packet.decode('ascii')
        logger.info("%s %s", addr, action)
        self._s.sendto(self._action(action).encode('ascii'), addr)
        return False

    def run(self) -> None:
        sel = selectors.DefaultSelector()
        sel.register(self._s, selectors.EVENT_READ, self._respond)
        sel.register(self._r, selectors.EVENT_READ, lambda: True)

        self.state = {
            'l-band': {
                'rf-ptt': False,
                'pa-power': False,
                'lna': False,
            },
            'uhf': {
                'rf-ptt': False,
                'pa-power': False,
                'lna': False,
            },
        }

        stop = False
        while not stop:
            for key, _ in sel.select():
                if stop := key.data():
                    break

        sel.close()
        self._s.close()
        os.close(self._r)
        os.close(self._w)
        logger.info("Stopped")

    def close(self) -> None:
        os.write(self._w, b's')


if __name__ == "__main__":
    Stationd(("127.0.0.2", 5005)).start()



#!/usr/bin/env python3

import argparse
import select
from cmd import Cmd
from pathlib import Path
from threading import Thread
from typing import Never

from pass_commander.config import AzEl
from pass_commander.rotator import Rotator


class RotCmd(Cmd):
    def __init__(self, rot: Rotator) -> None:
        '''Small cmdline to interact with a rotator.'''
        super().__init__()
        self.rot = rot
        Thread(target=self._listen, daemon=True).start()

    def _listen(self) -> Never:
        epoll = select.epoll()
        epoll.register(self.rot.listener, select.EPOLLIN)
        while True:
            for _ in epoll.poll(-1):
                print(f'\n{self.rot.event()}\n{self.prompt}', end='')

    def do_limits(self, _arg: str) -> None:
        print(*self.rot.limits())

    def do_go(self, arg: str) -> None:
        az, el = arg.split()
        pos = AzEl(float(az), float(el))
        self.rot.go(pos)
        self.rot.start_polling(pos)

    def do_pos(self, _arg: str) -> None:
        print(self.rot.position())

    def do_park(self, _arg: str) -> None:
        self.rot.park()
        self.rot.start_polling(AzEl(180, 90))

    def do_ppd(self, _arg: str) -> None:
        print(self.rot.ppd)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('path', type=Path, help='rotator /dev path')
    args = parser.parse_args()

    rot = Rotator(args.path)
    try:
        RotCmd(rot).cmdloop()
    finally:
        rot.close()



#!/usr/bin/env python3
"""Tiny testing script to send ascii hex bytes from stdin to the Pass Commander EDL socket."""

import socket
from argparse import ArgumentParser
from contextlib import suppress
from time import sleep


def _main() -> None:
    parser = ArgumentParser("Send ascii hex from stdin to the EDL socket")
    parser.add_argument(
        "-p",
        "--port",
        default=10025,
        type=int,
        help="port to use for the uplink, default is %(default)s",
    )
    args = parser.parse_args()

    packet = bytes.fromhex(input())

    edl = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    edl.connect(("127.0.0.1", args.port))

    while True:
        print("<--", packet.hex())
        edl.send(packet)
        sleep(1)


if __name__ == "__main__":
    with suppress(KeyboardInterrupt):
        _main()



# ruff: noqa: ERA001

import pytest
from skyfield.api import E, N, Time, wgs84
from skyfield.units import Angle, Velocity

from pass_commander.commander import Commander, SinglePass
from pass_commander.config import Config
from pass_commander.mock.flowgraph import Edl, Flowgraph
from pass_commander.mock.rotator import PtyRotator
from pass_commander.mock.station import Stationd
from pass_commander.satellite import Satellite
from pass_commander.tracker import Tracker


@pytest.fixture
def mock_config(
    good_config: Config,
    stationd: Stationd,
    rotator: PtyRotator,
    edl: Edl,
    flowgraph: Flowgraph,
) -> Config:
    '''Generate a config with all mock devices started.'''
    good_config.station = stationd
    good_config.rotator = rotator
    good_config.edl_dest = edl.addr
    good_config.flowgraph = flowgraph.addr
    return good_config


class TestSinglePass:
    def test_work_pass(self, mock_config: Config, sat: Satellite) -> None:
        sp = SinglePass(mock_config, lna_delay=0.0, morse_delay=0.0, cooloff_delay=0.0)
        sp.rot.cmd_interval = 0

        tk = Tracker(mock_config.observer)
        times = tk.next_pass(sat, sat.epoch)
        (pt, az, el), (rt, rv) = tk.track(sat, times)
        # now = tk.ts.now()
        # Start now
        # pt += now - pt[0]
        # rt += now - rt[0]
        #  1000x faster
        # pt *= 0.001
        # rt *= 0.001
        # FIXME: scale time to be quick
        pt -= 1000
        rt -= 1000

        sp.work((pt, az, el), (pt, az, el), (rt, rv))

    def test_over_thermal_limit(self, mock_config: Config, sat: Satellite) -> None:
        mock_config.temp_limit = 24.0
        sp = SinglePass(mock_config, lna_delay=0.0, morse_delay=0.0, cooloff_delay=0.0)
        sp.rot.cmd_interval = 0

        tk = Tracker(mock_config.observer)
        times = tk.next_pass(sat, sat.epoch)
        (pt, az, el), (rt, rv) = tk.track(sat, times)
        pt += 1000
        rt += 1000
        with pytest.raises(RuntimeError, match=r"^Temperature too high"):
            sp.work((pt, az, el), (pt, az, el), (rt, rv))


class TestCommander:
    def test_pointing_mode(self, mock_config: Config) -> None:
        class FakePass:
            def work(
                self,
                _pos: tuple[Time, Angle, Angle],
                nav: tuple[Time, Angle, Angle],
                _rv: tuple[Time, Velocity],
            ) -> None:
                self.target_az = nav[1].degrees[0]

        origin = wgs84.latlon(45 * N, -122 * E)
        mock_config.observer = origin

        cmdr = Commander(mock_config)
        cmdr.singlepass = FakePass()

        lat = origin.latitude.degrees
        lon = origin.longitude.degrees

        targets = [
            (wgs84.latlon(lat + 0.1, lon + 0.0), 0),  # North
            (wgs84.latlon(lat + 0.1, lon + 0.1), 45),  # NE
            (wgs84.latlon(lat + 0.1, lon - 0.1), 315),  # NW
            (wgs84.latlon(lat - 0.1, lon + 0.0), 180),  # South
            (wgs84.latlon(lat - 0.1, lon + 0.1), 135),  # SE
            (wgs84.latlon(lat - 0.1, lon - 0.1), 225),  # SW
            (wgs84.latlon(lat + 0.0, lon + 0.1), 90),  # East
            (wgs84.latlon(lat + 0.0, lon - 0.1), 270),  # West
        ]
        for target, expected_az in targets:
            cmdr.point(target)
            assert cmdr.singlepass.target_az == pytest.approx(expected_az)



import logging
from ctypes import c_char_p, cdll
from pathlib import PosixPath
from threading import Thread
from typing import Literal

import serial
from rot2prog import ROT2ProgSim


class PtyRotator(ROT2ProgSim):  # type: ignore[misc]
    def __init__(self, pulses_per_degree: Literal[1, 2, 4]) -> None:
        """Create object, open a serial connection, and start the daemon thread to run simulator.

        Args:
            pulses_per_degree (int): Resolution of simulated ROT2Prog controller. Options are 1,
            2, and 4.
        """
        self._log = logging.getLogger(__name__)

        # open serial port
        self._ser = serial.Serial(
            port='/dev/ptmx',
            baudrate=600,
            bytesize=8,
            parity='N',
            stopbits=1,
            timeout=None,
            # inter_byte_timeout allows continued operation after a bad packet
            inter_byte_timeout=0.1,
        )

        # FIXME: When Python 3.13 is standard, replace with os.{grantpt|unlockpt|ptsname}
        libc = cdll.LoadLibrary("libc.so.6")
        libc.grantpt(self._ser.fd)
        libc.unlockpt(self._ser.fd)
        libc.ptsname.restype = c_char_p
        self._client_path = PosixPath(libc.ptsname(self._ser.fd).decode('utf-'))

        self._pulses_per_degree = int(pulses_per_degree)
        self._log.info('ROT2Prog simulation interface opened on ' + str(self._ser.name))

        # start daemon thread to communicate on serial port
        # FIXME: daemon = True, removed for stop() testing
        Thread(target=self._run).start()

    @property
    def client_path(self) -> PosixPath:
        return self._client_path

    # FIXME: remove when upstream is updated
    def close(self) -> None:
        super().stop()
        self._ser.cancel_read()

    def _run(self) -> None:
        # Because upstream doesn't handle cancel_read it sometimes OSError: Errno 5
        try:
            super()._run()
        except OSError as e:
            # It's actually a SerialException but it doesn't set errno
            if str(e) != 'read failed: [Errno 5] Input/output error':
                raise
        self._log.info("Stopped")


if __name__ == '__main__':
    from time import sleep

    rot = PtyRotator(1)
    print(rot.client_path)  # noqa: T201
    try:
        while True:
            sleep(100)
    finally:
        rot.close()



from contextlib import closing

import pytest

from pass_commander.station import Station, StationError


class TestStation:
    def test_pa(self, stationd: tuple[str, int]) -> None:
        with closing(Station(stationd)) as s:
            s.pa_on()
            s.pa_off()

    def test_ptt(self, stationd: tuple[str, int]) -> None:
        with closing(Station(stationd)) as s:
            s.ptt_on()
            s.ptt_off()

    def test_getttemp(self, stationd: tuple[str, int]) -> None:
        with closing(Station(stationd)) as s:
            assert s.gettemp() == 25.0

    def test_lna(self, stationd: tuple[str, int]) -> None:
        with closing(Station(stationd, lna_delay=0)) as s:
            s.lna_on()
            s.lna_off()

    def test_pa_ptt_conflict(self, stationd: tuple[str, int]) -> None:
        with closing(Station(stationd)) as s:
            s.pa_on()
            s.ptt_on()
            with pytest.raises(StationError, match='PTT Conflict'):
                s.pa_off()

    def test_lna_ptt_conflict(self, stationd: tuple[str, int]) -> None:
        with closing(Station(stationd, lna_delay=0)) as s:
            s.ptt_on()
            with pytest.raises(StationError):
                s.lna_on()

    def test_invalid_station(self, stationd: tuple[str, int]) -> None:
        with (
            closing(Station(stationd, band='fakeband')) as s,
            pytest.raises(StationError, match='invalid command'),
        ):
            s.pa_on()



import selectors
from contextlib import closing
from math import isclose

import pytest

from pass_commander.config import AzEl
from pass_commander.mock.rotator import PtyRotator
from pass_commander.rotator import Rotator


class TestRotator:
    @pytest.fixture
    def rot(self, rotator: PtyRotator) -> Rotator:
        with closing(Rotator(rotator, cmd_interval=0)) as r:
            yield r

    def test_go(self, rot: Rotator) -> None:
        sel = selectors.DefaultSelector()
        sel.register(rot.listener, selectors.EVENT_READ, None)

        for pos in (
            AzEl(20, 20),
            AzEl(10.1, 10.4),
            AzEl(0, 0),
            AzEl(14.555346144996733, 0.4611240576248191),
        ):
            rot.go(pos)
            rot.start_polling(pos)
            sel.select()
            event = rot.event()
            assert isinstance(event, AzEl)
            # FIXME: see rotator.py _ppd comment. Remove 0.1 when fixed
            assert isclose(event.az, pos.az, rel_tol=0, abs_tol=1 / rot.ppd + 0.1)
            assert isclose(event.el, pos.el, rel_tol=0, abs_tol=1 / rot.ppd + 0.1)

    def test_double_go(self, rot: Rotator) -> None:
        rot.go(AzEl(0, 0))
        rot.go(AzEl(0, 0))

    def test_park(self, rot: Rotator) -> None:
        rot.park()
        assert rot.position() == AzEl(180, 90)

    def test_limits(self, rot: Rotator) -> None:
        az, el = rot.limits()
        good = [AzEl(az.lower, 0), AzEl(az.upper, 0), AzEl(0, el.lower), AzEl(0, el.upper)]
        bad = [
            AzEl(az.lower - 1, 0),
            AzEl(az.upper + 1, 0),
            AzEl(0, el.lower - 1),
            AzEl(0, el.upper + 1),
        ]

        for azel in good:
            rot.go(azel)
        for azel in bad:
            with pytest.raises(ValueError, match="out of range"):
                rot.go(azel)



import logging
from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from pathlib import Path
from textwrap import dedent

from skyfield.api import E, N, wgs84

from . import config, mock
from .commander import Commander

logger = logging.getLogger(__name__)


def handle_args() -> Namespace:  # noqa: D103
    parser = ArgumentParser(formatter_class=RawTextHelpFormatter)
    parser.add_argument(
        "-a",
        "--action",
        choices=("run", "dryrun", "nextpass"),
        help=dedent(
            """\
            Which action to have Pass Commander take
            - run: Normal operation
            - dryrun: Simulate the next pass immediately
            - nextpass: Sleep until next pass and then quit
            Default: '%(default)s'"""
        ),
        default="run",
    )
    parser.add_argument(
        "-c",
        "--config",
        default="~/.config/OreSat/pass_commander.toml",
        type=Path,
        help=dedent(
            """\
            Path to .toml config file. If dir will assume 'pass_commander.toml' in that dir
            Default: '%(default)s'"""
        ),
    )
    parser.add_argument(
        "--template",
        action="store_true",
        help="Generate a config template at the path specified by --config",
    )
    parser.add_argument(
        "-e",
        "--edl-port",
        type=int,
        default=10025,
        help="Port to listen for EDL packets on, default: %(default)s",
    )
    parser.add_argument(
        "-m",
        "--mock",
        action="append",
        choices=("tx", "rot", "con", "all"),
        help=dedent(
            """\
            Use a simulated (mocked) external dependency, not the real thing
            - tx: No PTT or EDL bytes sent to flowgraph
            - rot: No actual movement commanded for the rotator
            - con: Don't use network services - TLEs, weather, rot2prog, stationd
            - all: All of the above
            Can be issued multiple times, e.g. '-m tx -m rot' will disable tx and rotator"""
        ),
    )
    parser.add_argument(
        "--pass-count",
        type=int,
        default=9999,
        help="Maximum number of passes to operate before shutting down. Default: '%(default)s'",
    )
    parser.add_argument(
        "-s",
        "--satellite",
        help=dedent(
            """\
            Can be International Designator, Catalog Number, or Name.
            If `--mock con` is specified will search local TLE cache and Gpredict cache
            """
        ),
    )
    parser.add_argument(
        "-t", "--tx-gain", type=int, help="Transmit gain, usually between 0 and 100ish"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="Output additional debugging information",
    )
    parser.add_argument(
        "--temperature-limit",
        type=float,
        help="Temperature in Celsius of the station above which prevents a pass from running",
    )
    parser.add_argument(
        "-p",
        "--point",
        help="Point antenna at a given coordinate and start a pass. Format: <lat>,<lon> in decimal",
    )
    return parser.parse_args()


def _cfgerr(args: Namespace, msg: str) -> None:
    # This function is always called from an exception handler
    logger.debug("Config error", exc_info=True)  # noqa: LOG014
    logger.error("In '%s': %s", args.config, msg)


def main() -> None:  # noqa: D103 C901 PLR0912 PLR0915
    logging.basicConfig(
        level=logging.INFO, format='%(asctime)s [%(levelname)s] %(name)-25s: %(message)s'
    )

    args = handle_args()
    if args.config.is_dir():
        args.config /= "pass_commander.toml"

    if args.template:
        try:
            config.Config.template(args.config)
        except FileExistsError:
            _cfgerr(args, 'delete existing file before creating template')
        else:
            logger.info("Config template generated at '%s'", args.config)
            logger.info("Edit '%s' <template text> before running again", args.config)
        return

    try:
        conf = config.Config(args.config)
    except config.ConfigNotFoundError as e:
        _cfgerr(
            args,
            f"the file is missing ({type(e.__cause__).__name__}). Initialize using --template",
        )
    except config.InvalidTomlError as e:
        _cfgerr(args, f"there is invalid toml: {e}\nPossibly an unquoted string?")
    except config.MissingKeyError as e:
        _cfgerr(args, f"required key '{e.table}.{e.key}' is missing")
    except config.TemplateTextError as e:
        _cfgerr(args, f"key '{e}' still has template text. Replace <angle brackets>")
    except config.UnknownKeyError as e:
        _cfgerr(args, f"remove unknown keys: {' '.join(e.keys)}")
    except config.KeyValidationError as e:
        _cfgerr(args, f"key '{e.table}.{e.key}' has invalid type {e.actual}, expected {e.expect}")
    except config.IpValidationError as e:
        _cfgerr(args, f"contents of '{e.table}.{e.key}' is not a valid IP")
    except config.TleValidationError as e:
        _cfgerr(args, f"TLE '{e.name}' is invalid: {e.__cause__}")
    else:
        conf.mock = set(args.mock or [])
        if 'all' in conf.mock:
            conf.mock = {'tx', 'rot', 'con'}
        # Favor command line values over config file values
        conf.txgain = args.tx_gain or conf.txgain
        conf.sat_id = args.satellite or conf.sat_id
        if not conf.sat_id:
            logger.error(
                "No satellite specified. Set on command line (see --help) or in config file."
            )
            return
        conf.pass_count = args.pass_count
        conf.temp_limit = args.temperature_limit or conf.temp_limit

        mock_edl = None
        mock_flowgraph = None
        mock_stationd = None
        mock_rotator = None

        if 'con' in conf.mock:
            # Radio mock
            conf.edl = ("127.0.0.1", conf.edl[1])
            mock_edl = mock.Edl()
            conf.edl_dest = mock_edl.addr
            mock_edl.start()
            mock_flowgraph = mock.Flowgraph()
            conf.flowgraph = mock_flowgraph.addr
            mock_flowgraph.start()
            # Tracker mock
            conf.owmid = ''

        if 'tx' in conf.mock:
            mock_stationd = mock.Stationd()
            conf.station = mock_stationd.addr
            mock_stationd.start()

        if 'rot' in conf.mock:
            mock_rotator = mock.PtyRotator(pulses_per_degree=1)
            conf.rotator = mock_rotator.client_path

        commander = Commander(conf)

        try:
            if args.point is not None:
                lat, lon = args.point.split(',')
                commander.point(wgs84.latlon(float(lat) * N, float(lon) * E, 50))
            elif args.action == 'run':
                commander.autorun(count=conf.pass_count)
            elif args.action == 'dryrun':
                commander.dryrun()
            elif args.action == 'nextpass':
                sat, np = commander.sleep_until_next_pass()
                logger.info('Slept for pass %s by sat %s', np, sat)
            else:
                logger.info("Unknown action: %s", args.action)
        finally:
            # TODO: context manager/closeable for mocks?
            if mock_edl is not None:
                mock_edl.close()
            if mock_flowgraph is not None:
                mock_flowgraph.close()
            if mock_stationd is not None:
                mock_stationd.close()
            if mock_rotator is not None:
                mock_rotator.close()



from contextlib import closing
from itertools import pairwise

import pytest
from skyfield.api import E, N, wgs84

from pass_commander.mock import Edl, Flowgraph
from pass_commander.radio import Radio
from pass_commander.satellite import Satellite
from pass_commander.tracker import Tracker


class TestRadio:
    @pytest.fixture
    def radio(self, flowgraph: Flowgraph, edl: Edl) -> Radio:
        with closing(Radio(flowgraph.addr, edl.addr, "TEST")) as r:
            yield r

    def test_doppler(self, sat: Satellite, flowgraph: Flowgraph, edl: Edl) -> None:
        rx = []
        tx = []
        flowgraph._server.register_function(lambda x: rx.append(x), "set_gpredict_rx_frequency")  # noqa: SLF001
        flowgraph._server.register_function(lambda x: tx.append(x), "set_gpredict_tx_frequency")  # noqa: SLF001

        with closing(Radio(flowgraph.addr, edl.addr, "TEST")) as radio:
            tracker = Tracker(wgs84.latlon(45.509054 * N, -122.681394 * E, 50))
            nextpass = tracker.next_pass(sat, after=sat.epoch)
            _, (_, rangevel) = tracker.track(sat, nextpass)

            for v in rangevel.m_per_s:
                radio.set_rx_frequency(v)
                radio.set_tx_frequency(v)

        # RX doppler moves from high frequency to low frequency, passing over the center frequency
        assert rx[0] > radio.rxfreq > rx[-1]
        # monotonically decreasing
        assert all(x > y for x, y in pairwise(rx))

        # TX does the opposite, monotonically low -> high, again passing the center frequency
        assert tx[0] < radio.txfreq < tx[-1]
        assert all(x < y for x, y in pairwise(tx))

    def test_ident(self, radio: Radio) -> None:
        radio.morse_delay = 0
        radio.ident()

    def test_selector(self, radio: Radio) -> None:
        mode = 'cw'
        radio.set_tx_selector(mode)
        assert radio.get_tx_selector() == mode

    def test_tx_gain(self, radio: Radio) -> None:
        gain = 55
        radio.set_tx_gain(gain)

    def test_morse_ident(self, radio: Radio) -> None:
        ident = "OTHER"
        radio.set_morse_ident(ident)

        # FIXME: check flowgraph value

    def test_edl(self, radio: Radio) -> None:
        packet = "test string".encode('ascii')
        radio.edl(packet, 0)



from datetime import timedelta

import responses
from skyfield.api import E, N, wgs84

from pass_commander.satellite import Satellite
from pass_commander.tracker import Tracker


class TestTracker:
    def test_next_pass(self, sat: Satellite) -> None:
        track = Tracker(wgs84.latlon(45.509054 * N, -122.681394 * E, 50))
        np = track.next_pass(sat, after=sat.epoch)
        assert np.rise.time < np.culm[0].time < np.fall.time
        if np.fall.time - np.rise.time < (timedelta(hours=1) / timedelta(days=1)):
            track.track(sat, np)

    @responses.activate
    def test_weather(self) -> None:
        lat = 45.509054
        lon = -122.681394
        track = Tracker(wgs84.latlon(lat * N, lon * E, 50))
        # No owmid, returns default, should not make a request
        assert track.weather() == (25.0, 1010.0)

        # With owmid it should make a request
        track.owmid = "testid"
        weather = (2.0, 994.0)
        with responses.RequestsMock() as rsps:
            rsps.add(
                responses.GET,
                url='https://api.openweathermap.org/data/3.0/onecall',
                match=[
                    responses.matchers.query_param_matcher(
                        {"lat": f"{lat:.3f}", "lon": f"{lon:.3f}", "appid": track.owmid},
                        strict_match=False,
                    )
                ],
                body=f'{{"current": {{"temp": {weather[0]}, "pressure": {weather[1]}}}}}',
            )
            assert track.weather() == weather



from typing import Final

import pytest
import requests
import responses

from pass_commander.satellite import Satellite


class TestSatellite:
    # Args: sat_id, tle_cache, local_only
    # Sat ID types:
    # - 2024-149BK
    # - 60525
    # - ORESAT0.5
    # tle_cache some/none/sat not found
    # local_only t/f
    # celestrak success/fail/sat not found
    # how to test gpredict?
    # A sat not found via any method

    tle: Final[list[str]] = [
        'ORESAT0.5',
        '1 60525U 24149BK  25028.49486722  .00019661  00000+0  77233-3 0  9994',
        '2 60525  97.4192 108.1185 0006873 288.2875  71.7615 15.25712956 25052',
    ]

    names: Final[list[str]] = [
        '2024-149BK',
        '60525',
        'ORESAT0.5',
    ]

    @responses.activate
    def test_cache(self) -> None:
        cache = dict.fromkeys(self.names, self.tle)

        for name in self.names:
            Satellite(name, tle_cache=cache, local_only=True)

        with pytest.raises(ValueError, match=r"^Invalid satellite identifier"):
            Satellite('invalid', tle_cache=cache, local_only=True)

    @responses.activate
    def test_celestrak(self) -> None:
        # valid
        # missing
        # timeout
        # 403

        valid = responses.Response(
            method="GET",
            url=f"https://celestrak.org/NORAD/elements/gp.php?CATNR={self.names[1]}",
            body='\n'.join(self.tle),
        )
        missing = responses.Response(
            method="GET",
            url="https://celestrak.org/NORAD/elements/gp.php?NAME=missing",
            body='No GP data found',
        )
        forbidden = responses.Response(
            method="GET",
            url="https://celestrak.org/NORAD/elements/gp.php?NAME=forbidden",
            status=403,
        )
        responses.add(valid)
        responses.add(missing)
        responses.add(forbidden)

        Satellite(self.names[1])

        with pytest.raises(ValueError, match=r"^Invalid satellite identifier"):
            Satellite('missing')

        with pytest.raises(requests.exceptions.HTTPError):
            Satellite('forbidden')

    @responses.activate
    def test_cache_fallback(self) -> None:
        fallback = responses.Response(
            method="GET",
            url="https://celestrak.org/NORAD/elements/gp.php?NAME=fallback",
            body='No GP data found',
        )
        responses.add(fallback)

        Satellite('fallback', tle_cache={'fallback': self.tle})



import logging
import select
import socket
from datetime import timedelta
from functools import partial
from inspect import signature
from math import atan2, tau
from time import sleep

import linuxfd
import numpy as np
from jeepney import DBusAddress, Properties
from jeepney.io.blocking import open_dbus_connection
from skyfield.api import Time, load
from skyfield.toposlib import GeographicPosition
from skyfield.units import Angle, Velocity

from .config import AzEl, Config
from .radio import Radio
from .rotator import Rotator
from .satellite import Satellite
from .station import Station
from .tracker import PassInfo, Tracker

logger = logging.getLogger(__name__)


class SinglePass:
    def __init__(
        self,
        conf: Config,
        lna_delay: float | None = None,
        morse_delay: float | None = None,
        cooloff_delay: float | None = None,
    ) -> None:
        '''Assemble all things needed to run a single pass.'''
        # FIXME fetch these values from stationd/gnuradio
        if lna_delay is None:
            lna_delay = signature(Station.__init__).parameters['lna_delay'].default
        if morse_delay is None:
            morse_delay = signature(Radio.__init__).parameters['morse_delay'].default
        if cooloff_delay is None:
            cooloff_delay = 120.0
        self.conf = conf
        # TODO: check if resources are available but don't require until start of pass
        self.sta = Station(conf.station, band='l-band')
        self.uhf = Station(conf.station, band='uhf', lna_delay=lna_delay)
        self.rot = Rotator(conf.rotator, cal=conf.cal)
        self.rad = Radio(conf.flowgraph, conf.edl_dest, conf.name, morse_delay=morse_delay)
        self.ts = load.timescale()

        self.cooloff_delay = cooloff_delay
        self.min_el = 15  # FIXME: move to calc

    def work(
        self,
        pos: tuple[Time, Angle, Angle],
        nav: tuple[Time, Angle, Angle],
        rv: tuple[Time, Velocity],
    ) -> None:
        '''Do all actions for a single pass.

        Parameters
        ----------
        pos
            Position of the satellite, given in (time, azimuth, elevation)
        nav
            Navigation positions for the antenna, given in (time, azimuth, elevation). The rotator
            may be using an alternate track mode (e.g. flip, backhand) so az and el aren't always
            going to be the same as pos. Also it may have different times. El can go from 0 - 180.
        rv
            Range velocity of the satellite, as in how fast it's approaching or receding from the
            observer, given in (time, velocity). Used for doppler shift calculations.

        '''
        # Tasks:
        # - Pre-position antenna
        # - configure gain/radios
        # - start rx doppler
        # - at 0 degrees - send morse
        # - at min_el degrees - enable edl
        # - ability to mark maxel
        # - at min_el degrees - disable edl
        # - at 0 degrees - send morse
        # - park antenna
        # - disable gain/radios

        # calculate events array for each task?
        self.epoll = select.epoll()
        self.epoll.register(self.rot.listener, select.EPOLLIN)
        aosfd = linuxfd.timerfd(rtc=True, nonBlocking=True)
        losfd = linuxfd.timerfd(rtc=True, nonBlocking=True)
        rotfd = linuxfd.timerfd(rtc=True, nonBlocking=True)
        radfd = linuxfd.timerfd(rtc=True, nonBlocking=True)
        posfd = linuxfd.timerfd(rtc=True, nonBlocking=True)
        thmfd = linuxfd.timerfd(rtc=True, nonBlocking=True)

        # FIXME: log messages at AOS, Max el, LOS
        times, az, el = pos
        # first time above min_el
        aos = times[np.argmax(el.degrees > self.min_el)]
        # last time above min_el
        los = times[len(el.degrees) - 1 - np.argmax(np.flip(el.degrees) > self.min_el)]

        logger.info("AOS: %s", aos.utc_datetime())
        logger.info("LOS: %s", los.utc_datetime())

        self.action = {
            aosfd.fileno(): partial(self.on_rise, aosfd),
            losfd.fileno(): partial(self.on_fall, losfd),
            rotfd.fileno(): partial(
                self.on_rotator, rotfd, (iter(nav[0]), iter(nav[1].degrees), iter(nav[2].degrees))
            ),
            radfd.fileno(): partial(self.on_rx_doppler, radfd, (iter(rv[0]), iter(rv[1].m_per_s))),
            posfd.fileno(): partial(
                self.on_pos, posfd, (iter(times), iter(az.degrees), iter(el.degrees))
            ),
            thmfd.fileno(): partial(self.on_thermal, thmfd),
        }

        self.edl: socket.socket | None = None

        # Orient antenna to where the satellite wil rise
        # FIXME: compensate for slew rate, point at midpointish thing
        try:
            self.pre_position(nav[1].degrees[0], nav[2].degrees[0])

            self.epoll.register(aosfd.fileno(), select.EPOLLIN)
            self.epoll.register(losfd.fileno(), select.EPOLLIN)
            self.epoll.register(rotfd.fileno(), select.EPOLLIN)
            self.epoll.register(radfd.fileno(), select.EPOLLIN)
            self.epoll.register(posfd.fileno(), select.EPOLLIN)
            self.epoll.register(thmfd.fileno(), select.EPOLLIN)

            aosfd.settime(aos.utc_datetime().timestamp(), absolute=True)
            losfd.settime(los.utc_datetime().timestamp(), absolute=True)
            rotfd.settime(nav[0][0].utc_datetime().timestamp(), absolute=True)
            radfd.settime(rv[0][0].utc_datetime().timestamp(), absolute=True)
            posfd.settime(pos[0][0].utc_datetime().timestamp(), absolute=True)
            thmfd.settime(self.ts.now().utc_datetime().timestamp(), absolute=True)

            stop = False
            while not stop:
                for fd, event in self.epoll.poll(-1):
                    try:
                        logger.debug("%s", self.action[fd])
                        if stop := not self.action[fd](event):
                            break
                    except StopIteration:
                        self.epoll.unregister(fd)
        except Exception:
            # It'll take a while to get through the finally block so notify the user early on error
            logger.exception("!!Work pass interrupted:")
            raise
        finally:
            self.reset_hardware()

    def reset_hardware(self) -> None:
        logger.info("Pass ending, safing hardware")
        self.sta.ptt_off()
        if self.edl is not None:
            self.edl.close()
        self.uhf.lna_off()
        self.rad.set_tx_gain(3)

        self.rot.park()
        logger.info("Parked rotator")
        logger.info("Waiting %ds for PA to cool", self.cooloff_delay)
        sleep(self.cooloff_delay)
        self.sta.pa_off()

    def ident(self) -> bool:
        # The identifier must be sent
        # - Within 10 minutes of operating
        # - Every 10 minutes after that
        # - At the end of operating

        logger.info("Sending morse identifier")
        self.sta.ptt_on()
        try:
            self.rad.ident()
        finally:
            self.sta.ptt_off()
        logger.info("Morse identifier sent")
        return True

    def pre_position(self, az: float, el: float) -> None:
        self.rot.go(AzEl(az, el))
        logger.info("Started rotator movement")
        self.rot.start_polling(AzEl(az, el))
        # FIXME: guess from slew rate about how long it would take instead of
        # waiting indefinitely. We'll need some way of finding slew rate first
        # though.
        events = self.epoll.poll(-1)
        if len(events) > 1:
            raise RuntimeError("More events than expected")
        if events[0][0] != self.rot.listener:
            raise RuntimeError("Unexpected event fired")
        self.epoll.unregister(self.rot.listener)
        pos = self.rot.event()  # clears the event
        logger.info("pre-position arrived at %s", pos)
        self.on_positioned(events[0][1])

    def on_positioned(self, _event: int) -> bool:
        logger.info("Rotator at initial position, enabling pa, lna")
        self.rad.set_tx_gain(self.conf.txgain)
        self.uhf.lna_on()
        self.sta.pa_on()
        self.ident()
        logger.info("Initial position tasks complete")
        return True

    def on_rotator(
        self, timer: linuxfd.timerfd, nav: tuple[Time, Angle, Angle], _event: int
    ) -> bool:
        timer.read()
        time, az, el = next(nav[0]), next(nav[1]), next(nav[2])
        while time < self.ts.now():
            logger.info(
                '%-28s%s %7.3f°az %7.3f°el',
                'Skipping nav',
                time.utc_datetime(),
                az,
                el,
            )
            time, az, el = next(nav[0]), next(nav[1]), next(nav[2])
        timer.settime(time.utc_datetime().timestamp(), absolute=True)
        self.rot.go(AzEl(az, el))
        return True

    def on_pos(self, timer: linuxfd.timerfd, pos: tuple[Time, Angle, Angle], _event: int) -> bool:
        timer.read()
        time, az, el = next(pos[0]), next(pos[1]), next(pos[2])
        logger.info('%-28s: %7.3f°az %7.3f°el', "Satellite position", az, el)
        timer.settime(time.utc_datetime().timestamp(), absolute=True)
        return True

    def on_rise(self, timer: linuxfd.timerfd, _event: int) -> bool:
        logger.info("AOS %s", timer.read())
        self.rad.set_tx_selector("edl")
        self.sta.ptt_on()

        self.edl = socket.socket(socket.AF_INET, socket.SOCK_DGRAM | socket.SOCK_NONBLOCK)
        self.edl.bind(self.conf.edl)
        logger.info("EDL socket open")
        self.epoll.register(self.edl.fileno(), select.EPOLLIN | select.EPOLLERR)
        self.action[self.edl.fileno()] = partial(self.on_edl, self.edl)
        return True

    # FIXME: setup on_fall in on_rise so that self.edl can be local
    def on_fall(self, timer: linuxfd.timerfd, _event: int) -> bool:
        logger.info("LOS %s", timer.read())
        if self.edl is not None:
            self.edl.close()
        self.edl = None
        logger.info("EDL socket closed")
        self.ident()
        self.sta.ptt_off()
        return False

    def on_edl(self, edl: socket.socket, _event: int) -> bool:
        packet = edl.recv(4096)
        # FIXME: Recalculate current range_vel for exact time?
        self.rad.edl(packet, self.current_rv)
        logger.info("Sent EDL")
        return True

    def on_rx_doppler(self, timer: linuxfd.timerfd, rv: Velocity, _event: int) -> bool:
        timer.read()
        time, range_velocity = next(rv[0]), next(rv[1])
        while time < self.ts.now():
            logger.info(
                '%-28s%s %7.3f rv',
                'Skipping doppler',
                time.utc_datetime(),
                range_velocity,
            )
            time, range_velocity = next(rv[0]), next(rv[1])

        timer.settime(time.utc_datetime().timestamp(), absolute=True)
        # FIXME: set doppler to point that minimizes error over the interval (midpoint?)
        self.current_rv = range_velocity
        logger.debug("doppler %f", self.current_rv)
        self.rad.set_rx_frequency(self.current_rv)
        return True

    def on_thermal(self, timer: linuxfd.timerfd, _event: int) -> bool:
        timer.read()

        degc = self.sta.gettemp()
        if degc > self.conf.temp_limit:
            logger.info(
                "Temperature is too high (%.1f°C > %.1f°C). Skipping this pass.",
                degc,
                self.conf.temp_limit,
            )
            raise RuntimeError("Temperature too high")
        timer.settime(self.ts.now().utc_datetime().timestamp() + 30, absolute=True)
        return True


class Commander:
    def __init__(self, conf: Config) -> None:
        '''Create the main pass coordinator.'''
        self.conf = conf
        self.track = Tracker(conf.observer, owmid=conf.owmid)
        self.singlepass = SinglePass(conf)

    def require_clock_sync(self) -> None:
        def ntp_synchronized() -> bool:
            # FIXME: Paraphrased from man org.freedesktop.timedate1:
            # NTPSynchronized shows whether the kernel reports the time as
            # synchronized, reported by the system call adjtimex(3). The purpose of
            # this D-Bus property is to allow remote clients to access this
            # information. Local clients can access the information directly.
            #
            # I'd prefer to not use D-Bus but I can't find any existing Python
            # bindings for adjtimex() and the struct argument is sufficiently
            # complicated that I don't really want to write my own ctypes binding.
            msg = Properties(
                DBusAddress(
                    object_path='/org/freedesktop/timedate1',
                    bus_name='org.freedesktop.timedate1',
                    interface='org.freedesktop.timedate1',
                )
            ).get("NTPSynchronized")
            with open_dbus_connection(bus='SYSTEM') as con:
                return bool(con.send_and_get_reply(msg).body[0][1])

        while not ntp_synchronized():
            logger.warning("System clock is not synchronized. Sleeping 60 seconds.")
            sleep(60)
        logger.info("System clock is synchronized.")

    def sleep_until_next_pass(self) -> tuple[Satellite, PassInfo]:
        # FIXME: How many days is a TLE good for? Set next_pass(lookahead) based on that
        # FIXME: How often are TLEs updated? Set tle_refresh based on that
        # FIXME: Now that we have an event loop can this be part of it?

        while True:
            sat = Satellite(
                self.conf.sat_id, tle_cache=self.conf.tle_cache, local_only='con' in self.conf.mock
            )
            np = self.track.next_pass(sat, min_el=self.conf.min_el)
            if np is None:
                # FIXME: sleep until next TLE?
                raise RuntimeError("No pass found")

            to_sleep = timedelta(days=np.rise.time - self.track.ts.now())
            if to_sleep <= timedelta():
                logger.info("In a pass now!")
                return sat, np

            logger.info(
                "Sleeping %s until next rise time %s for a %.2f°el pass.",
                to_sleep,
                np.rise.time.astimezone(None),
                np.culm[0].el.degrees,
            )

            tle_refresh = timedelta(days=1)
            # FIXME: wake up a bit before a pass to adjust TLEs?
            sleep(min(to_sleep.total_seconds(), tle_refresh.total_seconds()))

            tle_age = timedelta(days=self.track.ts.now() - sat.epoch)
            if tle_age <= tle_refresh:
                return sat, np
            logger.info('TLE out of date by %s, refreshing', tle_age)

        raise RuntimeError("Unreachable?")

    def autorun(self, count: int) -> None:
        logger.info("Running for %d passes", count)
        while count > 0:
            self.require_clock_sync()
            sat, np = self.sleep_until_next_pass()

            # Pre-compute pass time/alt/az/rv
            temp, pressure = self.track.weather()
            logger.info("Current weather: %f°C %f mBar", temp, pressure)

            pos, rv = self.track.track(sat, np, temp, pressure)
            nav = self.singlepass.rot.path(np, pos)
            self.singlepass.work(pos, nav, rv)

            seconds = timedelta(days=np.fall.time - self.track.ts.now()).total_seconds()
            if seconds > 0:
                logger.info("Sleeping %.3f seconds until pass is really over.", seconds)
                sleep(seconds)
            count -= 1

    # Testing stuff goes below here

    def point(self, coord: GeographicPosition) -> None:
        # The real proper way of doing heading on a sphere involves the haversine formula but I
        # don't really want to implement it myself and also pull in a library just for this. Using
        # arctan assumes a flat plane but I claim (without having done any math whatsoever) that
        # this is sufficient since we're not intending to point at points past the horizon and also
        # not use this in the arctic circle.
        # FIXME: before we deploy a uniclogs station in Svalbard (one day I promise)
        n = coord.latitude.radians - self.conf.observer.latitude.radians
        e = coord.longitude.radians - self.conf.observer.longitude.radians
        # For trig functions like atan2 we treat N as x and E as y. It also outputs in the -180 -
        # 180 range but we want 0 - 360.
        az = atan2(e, n) % tau

        # A simulated pointing pass is two points, one at now with the appropriate azimuth, and one
        # forever far in the future that won't ever reasonably be reached.
        now = self.track.ts.now()
        # Python can't represent dates more than 10,000 years in the future and timerfds don't do
        # 10,000 days(?) so 1,000 days is now forever.
        times = self.track.ts.linspace(now, now + 1000, 2)
        pos = (times, Angle(radians=[az, az]), Angle(degrees=[0, 0]))
        nav = (times, Angle(radians=[az, az]), Angle(degrees=[0, 0]))
        rv = (times, Velocity.m_per_s([0, 0]))

        self.singlepass.work(pos, nav, rv)

    def dryrun(self) -> None:
        sat = Satellite(
            self.conf.sat_id, tle_cache=self.conf.tle_cache, local_only='con' in self.conf.mock
        )
        np = self.track.next_pass(sat, min_el=Angle(degrees=80), lookahead=timedelta(days=30))
        if np is None:
            raise RuntimeError("No pass found")
        logger.info(
            "Rise time: %s Max el: %.3f", np.rise.time.utc_datetime(), np.culm[0].el.degrees
        )
        (postime, az, el), (rvtime, rangevel) = self.track.track(sat, np)
        navtime, navaz, navel = self.singlepass.rot.path(np, (postime, az, el))
        offset = postime[0] - self.track.ts.now()
        logger.info("Dryrun offset: %f", offset)
        self.singlepass.work(
            (postime - offset, az, el),
            (navtime - offset, navaz, navel),
            (rvtime - offset, rangevel),
        )
        logger.info("Dry run complete")

    def test_morse(self) -> None:
        while True:
            self.singlepass.ident()
            sleep(30)
